# -*- coding: utf-8 -*-
"""
Test Scrapbag files file
"""
import time
import unittest
import mock

from scrapbag.tests.files.mock_requests import MockRequestResponse
from scrapbag.files import (
    open_filezip,
    extract_filezip,
    download_file,
    open_remote_url,
    is_remote_file_modified,
    copy_remote_file,
    remove_file
    )


class UtilsFilesTestCase(unittest.TestCase):
    """
    Scrapbag files Test CaseW
    """

    url = "http://test_url.com/test.csv"

    @mock.patch('zipfile.ZipFile')
    @mock.patch('zipfile.is_zipfile')
    def test_openzipfile(self, mock_iszip, mock_zipfile):
        """
        Test open_zipfile
        """
        mock_zipfile.return_value.configure_mock(**{
            'infolist.return_value': ['file1.txt', 'file2.csv', 'file2.txt'],
            'open.return_value': True,
        })
        mock_iszip.return_value = 'True'
        files = list(open_filezip('test.zip', '2'))

        mock_iszip.return_value = False
        files2 = list(open_filezip('test.meme', '2'))

        self.assertEqual(mock_iszip.call_count, 2)
        self.assertEqual(len(files), 2)
        self.assertEqual(len(files2), 0)

    @mock.patch('zipfile.ZipFile')
    @mock.patch('scrapbag.files.open', create=True)
    def test_extract_filezip(self, mock_open, mock_zipfile):
        """
        Test extract_filezip
        """

        files = ['file1.txt', 'file2.csv', 'file2.txt']

        mock_zipfile.configure_mock(**{
            'return_value.__enter__.return_value.namelist.return_value': files,
        })

        # extract only csv target
        result1 = extract_filezip('test.zip', 'dest_path', [r'.*\.csv'])

        # extract not existing target
        result2 = extract_filezip('test.zip', 'dest_path', [r'.*\.png'])

        # extract not existing target
        result3 = extract_filezip(
            'test.zip', 'dest_path', [r'.*\.png', r'.*\.csv'])

        # extract all without target
        result4 = extract_filezip('test.zip', 'dest_path', [])

        # extract all with target
        result5 = extract_filezip(
            'test.zip', 'dest_path', [r'.*\.csv', r'.*\.txt'])

        # force exception during extraction
        mock_zipfile__enter = mock_zipfile.return_value.__enter__
        mock_zipfile__enter.return_value.extract.side_effect = [
            True, Exception('Error extracting file'), False]
        result6 = extract_filezip(
            'test.zip', 'dest_path', [r'.*\.csv', r'.*\.txt'])

        result7 = extract_filezip('badzip.bad', 'dest_path')

        self.assertEqual(len(result1), 1)
        self.assertEqual(isinstance(result1, list), True)
        self.assertEqual(mock_open.call_count, 6)
        self.assertEqual(mock_open.return_value.close.call_count, 6)

        self.assertEqual(result2, [])
        self.assertEqual(result3, result1)

        self.assertEqual(result4, ['dest_path/%s' % (x,) for x in files])

        self.assertEqual(result5, result4)

        self.assertEqual(len(result6), 2)

        self.assertEqual(result7, [])

    @mock.patch('scrapbag.files.is_remote_file_modified')
    @mock.patch('scrapbag.files.copy_remote_file')
    @mock.patch('scrapbag.files.open_remote_url')
    def test_download_file(self, mock_open, mock_copy, mock_modified):
        """
        Test download_file
        """
        destination = "test_path/test_destination/"

        mock_open.return_value.configure_mock(**{
            'url': self.url,
            'close.return_value': True
        })

        mock_modified.return_value = True

        download_file(self.url, destination)
        self.assertTrue(mock_open.called)
        self.assertTrue(mock_modified.called)
        self.assertTrue(mock_copy.called)

        mock_copy.reset_mock()
        mock_open.reset_mock()
        mock_modified.reset_mock()

        mock_modified.return_value = False
        download_file(self.url, destination)

        self.assertFalse(mock_copy.called)
        self.assertTrue(mock_open.called)
        self.assertTrue(mock_modified.called)

        mock_open.return_value = False
        mock_modified.reset_mock()
        download_file(self.url, destination)

        self.assertFalse(mock_modified.called)

    @mock.patch('scrapbag.files.requests')
    def test_open_remote_url(self, mock_requests):
        """
        Test open_remote_url
        """

        mock_requests.get.return_value = MockRequestResponse(
            "ok", headers={'content-type': 'x-form-test'})
        result = open_remote_url(self.url)
        result2 = open_remote_url([self.url])

        self.assertIsNot(result, None)
        self.assertIsNot(result2, None)

        self.assertEqual(mock_requests.get.call_count, 2)

        mock_requests.get.return_value = MockRequestResponse(
            "ok", headers={'content-type': 'html'})
        result3 = open_remote_url([self.url])
        self.assertEqual(result3, None)

        mock_requests.side_effect = Exception('Test exception')
        result4 = open_remote_url([self.url])
        self.assertEqual(result4, None)

    @mock.patch('scrapbag.files.os.path.getmtime')
    @mock.patch('scrapbag.files.os.path.getsize')
    @mock.patch('scrapbag.files.os.path.exists')
    @mock.patch('scrapbag.files.time.gmtime')
    def test_is_remote_file_modified(self,
                                     mock_time,
                                     mock_exists,
                                     mock_getsize,
                                     mock_getmtime):
        """
        Test is_remote_file_modified
        """

        web_file_test = MockRequestResponse("ok", headers={
            'last-modified': 'Thu, 15 Sep 2016 12:03:01 GMT',
            'content-length': 123})

        test_time = time.strptime(
            web_file_test.headers['last-modified'], '%a, %d %b %Y %H:%M:%S %Z')

        mock_exists.return_value = True
        mock_getmtime.return_value = web_file_test.headers['last-modified']
        mock_getsize.return_value = web_file_test.headers['content-length']
        mock_time.return_value = test_time

        # no changed
        result = is_remote_file_modified(
            web_file_test, 'destination/test/file.csv')

        # change the last_modify param
        web_file_test.headers[
            'last-modified'] = 'Thu, 22 Sep 2016 12:03:01 GMT'
        result2 = is_remote_file_modified(
            web_file_test, 'destination/test/file.csv')

        # change the last_modify param & content_length
        web_file_test.headers['content-length'] = 456
        result3 = is_remote_file_modified(
            web_file_test, 'destination/test/file.csv')

        # change only content-length
        web_file_test.headers[
            'last-modified'] = 'Thu, 15 Sep 2016 12:03:01 GMT'
        web_file_test.headers['content-length'] = 456
        result4 = is_remote_file_modified(
            web_file_test, 'destination/test/file.csv')

        # destination doesnt exists
        mock_exists.return_value = False
        result5 = is_remote_file_modified(
            web_file_test, 'destination/test/file.csv')

        # no last_modified in headers
        web_file_test.headers['last-modified'] = None
        result6 = is_remote_file_modified(
            web_file_test, 'destination/test/file.csv')

        web_file_test.headers[
            'last-modified'] = 'Thu, 15 Sep 2016 12:03:01 GMT'
        web_file_test.headers['content-length'] = None
        result7 = is_remote_file_modified(
            web_file_test, 'destination/test/file.csv')

        mock_exists.side_effect = Exception('Test exception')
        result8 = is_remote_file_modified(
            web_file_test, 'destination/test/file.csv')

        self.assertEqual(result, False)
        self.assertEqual(result2, True)
        self.assertEqual(result3, True)
        self.assertEqual(result4, True)
        self.assertEqual(result5, True)
        self.assertEqual(result6, True)
        self.assertEqual(result7, True)
        self.assertEqual(result8, True)

    @mock.patch('scrapbag.files.open', create=True)
    @mock.patch('scrapbag.files.os.makedirs')
    @mock.patch('scrapbag.files.os.path.exists')
    def test_copy_remote_file(self, mock_exists, mock_makedirs, mock_open):
        """
        Test copy_remote_file
        """

        class WebFile:
            """
            Web file testing class
            """

            def __init__(self, value, **kwargs):
                """
                init webfile method
                """
                self.value = value
                self.stored = iter(value)

            def iter_content(self, **kwargs):
                """
                iter content method
                """
                return self.stored

            def reset(self):
                """
                reset method
                """
                self.stored = iter(self.value)

        # TODO: Refactor Webfile class.
        test_web_file = WebFile(['2', '3', None])

        mock_exists.return_value = True
        mock_makedirs.return_value = True
        mock_open.return_value = mock.MagicMock()

        copy_remote_file(test_web_file, 'test_path/path/file.csv')
        self.assertFalse(mock_makedirs.called)

        test_web_file.reset()
        mock_exists.return_value = False
        copy_remote_file(test_web_file, 'test_path/path/file.csv')
        self.assertTrue(mock_makedirs.called)

    @mock.patch('scrapbag.files.os.remove')
    @mock.patch('scrapbag.files.os.path.exists')
    def test_remove_file(self, mock_exists, mock_remove):
        """
        Test remove_file
        """
        mock_exists.return_value = True
        path = "test_path/dest/file.csv"

        remove_file(path)
        self.assertTrue(mock_remove.called)

        mock_remove.reset_mock()
        mock_exists.return_value = False
        remove_file(path)
        self.assertFalse(mock_remove.called)



# -*- coding: utf-8 -*-
"""
Scrapbag pdf file.
"""
import os
import logging
from io import StringIO

import pdfminer.layout
import pdfminer.settings
import pdfminer.high_level


import structlog

from .collections import exclude_empty_values

pdfminer.settings.STRICT = False
logging.getLogger('pdfminer').setLevel(logging.ERROR)

logger = structlog.getLogger(__name__)


def pdf_to_text(pdf_filepath='', **kwargs):
    """
    Parse pdf to a list of strings using the pdfminer lib.

    Args:
        no_laparams=False,
        all_texts=None,
        detect_vertical=None, word_margin=None, char_margin=None,
        line_margin=None, boxes_flow=None, codec='utf-8',
        strip_control=False, maxpages=0, page_numbers=None, password="",
        scale=1.0, rotation=0, layoutmode='normal', debug=False,
        disable_caching=False,
    """

    result = []
    try:
        if not os.path.exists(pdf_filepath):
            raise ValueError("No valid pdf filepath introduced..")

        # TODO: REVIEW THIS PARAMS
        # update params if not defined
        kwargs['outfp'] = kwargs.get('outfp', StringIO())
        kwargs['laparams'] = kwargs.get('laparams', pdfminer.layout.LAParams())
        kwargs['imagewriter'] = kwargs.get('imagewriter', None)
        kwargs['output_type'] = kwargs.get('output_type', "text")
        kwargs['codec'] = kwargs.get('codec', 'utf-8')
        kwargs['disable_caching'] = kwargs.get('disable_caching', False)

        with open(pdf_filepath, "rb") as f_pdf:
            pdfminer.high_level.extract_text_to_fp(f_pdf, **kwargs)

        result = kwargs.get('outfp').getvalue()

    except Exception:
        logger.error('fail pdf to text parsing')

    return result


def pdf_row_limiter(rows, limits=None, **kwargs):
    """
    Limit row passing a value. In this case we dont implementate a best effort
    algorithm because the posibilities are infite with a data text structure
    from a pdf.
    """
    limits = limits or [None, None]

    upper_limit = limits[0] if limits else None
    lower_limit = limits[1] if len(limits) > 1 else None

    return rows[upper_limit: lower_limit]


def pdf_row_format(row_str, **kwargs):
    """
    Retrieve formated pdf row by default as string.
    """
    return row_str


def pdf_row_parser(rows, **kwargs):
    """
    Retrieve rows parsed by default.
    """
    return rows


def pdf_row_cleaner(rows, **kwargs):
    """
    Check if there are any empty dict in the list of rows.
    """
    return exclude_empty_values(rows)


def pdf_to_dict(pdf_filepath, **kwargs):

    """
    Main method to parse a pdf file to a dict.
    """

    callbacks = {
        'pdf_to_text': pdf_to_text,
        'pdf_row_format': pdf_row_format,
        'pdf_row_limiter': pdf_row_limiter,
        'pdf_row_parser': pdf_row_parser,
        'pdf_row_cleaner': pdf_row_cleaner
        }

    callbacks.update(kwargs.get('alt_callbacks', {}))
    rows = kwargs.get('rows', [])

    if not rows:
        # pdf to string
        rows_str = callbacks.get('pdf_to_text')(pdf_filepath, **kwargs)

        # string to list of rows
        rows = callbacks.get('pdf_row_format')(rows_str, **kwargs)

    # apply limits
    rows = callbacks.get('pdf_row_limiter')(rows, **kwargs)

    # Parse data from rows to dict
    rows = callbacks.get('pdf_row_parser')(rows, **kwargs)

    # apply cleaner
    rows = callbacks.get('pdf_row_cleaner')(rows)

    return rows



# -*- coding: utf-8 -*-
"""
Datary MockRequestsResponse Test
"""
import unittest
from scrapbag.tests.files.mock_requests import MockRequestResponse


class MockRequestsResponseTestCase(unittest.TestCase):
    """
    MockRequestsResponse Test Case
    """

    def test(self):
        """
        Test MockRequestsResponse
        """
        test = MockRequestResponse(
            'aaaa', path='test_path', json={'test': [1, 2, 3]})
        self.assertEqual(test.text, 'aaaa')
        self.assertEqual(test.path(), 'test_path')
        self.assertEqual(test.encoding(), 'utf-8')
        self.assertEqual(test.json(), {'test': [1, 2, 3]})



# -*- coding: utf-8 -*-
"""
Scrapbag strings file.
"""
import re
import unicodedata
import html2text
import structlog
from bs4 import BeautifulSoup
from markdown import markdown

logger = structlog.getLogger(__name__)


def clean_text(text):
    """
    Retrieve clean text without markdown sintax or other things.
    """
    if text:
        text = html2text.html2text(clean_markdown(text))
        return re.sub(r'\s+', ' ', text).strip()


def clean_markdown(text):
    """
    Parse markdown sintaxt to html.
    """
    result = text

    if isinstance(text, str):
        result = ''.join(
            BeautifulSoup(markdown(text), 'lxml').findAll(text=True))

    return result


def get_only_words(string):
    """
    Get only words scapping numbers and simbols.
    """
    return re.findall(r"[\w']+", string)


# def strip_accents(s):
#     """
#     Change char with different locale with similar.
#     """
# return ''.join(c for c in unicodedata.normalize('NFD', s) if
# unicodedata.category(c) != 'Mn')

def select_regexp_char(char):
    """
    Select correct regex depending the char
    """
    regexp = '{}'.format(char)

    if not isinstance(char, str) and not isinstance(char, int):
        regexp = ''

    if isinstance(char, str) and not char.isalpha() and not char.isdigit():
        regexp = r"\{}".format(char)

    return regexp


def exclude_chars(text, exclusion=None):
    """
    Clean text string of simbols in exclusion list.
    """
    exclusion = [] if exclusion is None else exclusion
    regexp = r"|".join([select_regexp_char(x) for x in exclusion]) or r''
    return re.sub(regexp, '', text)


def strip_accents(text):
    """
    Strip agents from a string.
    """

    normalized_str = unicodedata.normalize('NFD', text)

    return ''.join([
        c for c in normalized_str if unicodedata.category(c) != 'Mn'])


OPERATIONS_EXCLUSION = ['+', '*', '%', '_', '//']


def normalizer(text, exclusion=[], lower=True, separate_char='-', **kwargs):
    """
    Clean text string of simbols only alphanumeric chars.
    """
    exclusion = OPERATIONS_EXCLUSION if not exclusion else exclusion
    clean_str = re.sub(r'[^\w{}]'.format(
        "".join(exclusion)), separate_char, text.strip()) or ''
    clean_lowerbar = clean_str_without_accents = strip_accents(clean_str)

    if '_' not in exclusion:
        clean_lowerbar = re.sub(r'\_', separate_char, clean_str_without_accents.strip())

    limit_guion = re.sub(r'\-+', separate_char, clean_lowerbar.strip())

    # TODO: refactor with a regexp
    if limit_guion and separate_char and separate_char in limit_guion[0]:
        limit_guion = limit_guion[1:]

    if limit_guion and separate_char and separate_char in limit_guion[-1]:
        limit_guion = limit_guion[:-1]

    if lower:
        limit_guion = limit_guion.lower()

    return limit_guion


def normalize_dict(dictionary, **kwargs):
    """
    Given an dict, normalize all of their keys using normalize function.
    """
    result = {}
    if isinstance(dictionary, dict):
        keys = list(dictionary.keys())
        for key in keys:
            result[normalizer(key, **kwargs)] = normalize_dict(dictionary.get(key), **kwargs)
    else:
        result = dictionary
    return result



# -*- coding: utf-8 -*-
"""
Test Scrapbag spss file
"""
import os

import mock
import unittest

from scrapbag.spss import (
    parse_spss_header_leyend,
    parse_variable_label,
    parse_value_label,
    parse_spss_headerfile,
    parse_spss_datafile,
    join_spss_header_data,
    parse_spss,
    )


class UtilsSpssTestCase(unittest.TestCase):

    def setUp(self):
        """
        Setup 3 test case from cis with differents formats to test parser.
        """
        self.SPSS_ROOT_PATH = 'scrapbag/tests/files/spss'

        self.spss_path1 = os.path.join(self.SPSS_ROOT_PATH, 'MD3193_2017')
        self.spss_path2 = os.path.join(self.SPSS_ROOT_PATH, 'MD2967_2012')
        self.spss_path3 = os.path.join(self.SPSS_ROOT_PATH, 'MD2564_2004')

        self.header_filepath3 = os.path.join(self.spss_path1, 'ES3193')
        self.data_filepath3 = os.path.join(self.spss_path1, 'DA3193')

        self.header_filepath2 = os.path.join(self.spss_path2, 'ES2967')
        self.data_filepath2 = os.path.join(self.spss_path2, 'DA2967')

        self.header_filepath1 = os.path.join(self.spss_path3, 'ES2564')
        self.data_filepath1 = os.path.join(self.spss_path3, 'DA2564')

    @mock.patch('scrapbag.spss.add_element')
    def test_parse_spss_header_leyend(self, mock_add_element):

        mock_add_element.side_effect = Exception('Test Exception')
        self.assertEqual(parse_spss_header_leyend(''), {})

    @mock.patch('scrapbag.spss.add_element')
    def test_parse_variable_label(self, mock_add_element):
        header = {'a': 123}
        mock_add_element.side_effect = Exception('Test Exception')
        self.assertEqual(parse_variable_label(["/P1401 'Hasta ahora'\r\n/P1401 'Hasta ahora'"], header), None)
        self.assertEqual(header, {'a': 123})

    def test_parse_value_label(self):

        pass

    def test_parse_spss_header_labels(self):

        pass

    def test_parse_spss_headerfile(self):

        pass

    def test_parse_spss_header_values(self):

        pass

    def test_parse_spss_datafile(self):

        pass

    def test_join_spss_header_data(self):

        pass

    def test_parse_spss_header_data_file(self):

        header1 = parse_spss_headerfile(self.header_filepath1)
        header2 = parse_spss_headerfile(self.header_filepath2)
        header3 = parse_spss_headerfile(self.header_filepath3)

        with mock.patch('scrapbag.spss.parse_spss_header_leyend') as mock_parse_leyend:
            mock_parse_leyend.return_value = None
            bad_header4 = parse_spss_headerfile(self.header_filepath1)
            self.assertEqual(bad_header4, {})

        # previus reviewed the header file
        self.assertEqual(len(header1), 62)
        self.assertEqual(len(header2), 172)
        self.assertEqual(len(header3), 162)

        data1 = parse_spss_datafile(self.data_filepath1)
        data2 = parse_spss_datafile(self.data_filepath2)
        data3 = parse_spss_datafile(self.data_filepath3)

        # previus reviewed the data files
        self.assertEqual(len(data1), 4781)
        self.assertEqual(len(data2), 2464)
        self.assertEqual(len(data3), 1510)

        result1 = join_spss_header_data(header1, data1)
        result2 = join_spss_header_data(header2, data2)
        result3 = join_spss_header_data(header3, data3)
        result4 = parse_spss(
            header_filepath=self.header_filepath1,
            data_filepath=self.data_filepath1)

        self.assertEqual(len(result1), 4781)
        self.assertEqual(result1, result4)
        self.assertEqual(len(result2), 2464)
        self.assertEqual(len(result3), 1510)



# -*- coding: utf-8 -*-
""" Reusable utils module."""

__all__ = [
    "collections",
    "csvs",
    "dates",
    "files",
    "geo",
    "pdf",
    "spss",
    "strings",
    "version"]

# BAD PRACTICE: the wildcard is bad but usefull.
from .collections import *
from .csvs import *
from .dates import *
from .files import *
from .geo import *
from .pdf import *
from .spss import *
from .strings import *
from .version import *



# -*- coding: utf-8 -*-
"""
Test Scrapbag pdf file
"""
import unittest
import mock

from scrapbag.pdf import (
    pdf_to_text,
    pdf_row_limiter,
    pdf_row_format,
    pdf_row_parser,
    pdf_row_cleaner,
    pdf_to_dict
    )


class UtilsPDFTestCase(unittest.TestCase):
    """
    Test Scrapbag pdf test case
    """

    test_rows = [
        ['test row 1 - aa', 'test row 1 - ab', 'test row 1 - ac'],
        ['test row 2 - ba', 'test row 2 - bb', 'test row 2 - bc'],
        ['test row 3 - ca', 'test row 3 - cb', 'test row 3 - cc'],
        ['test row 4 - da', 'test row 4 - db', 'test row 4 - dc'],
        ['test row 5 - ea', 'test row 5 - eb', 'test row 5 - ec']]

    @mock.patch('scrapbag.pdf.os')
    @mock.patch('scrapbag.pdf.open')
    @mock.patch('scrapbag.pdf.pdfminer.high_level')
    @mock.patch('scrapbag.pdf.StringIO')
    def test_pdf_to_text(self,
                         mock_stringio,
                         mock_pdfminer_highlevel,
                         mock_open,
                         mock_os):

        """
        Test pdf_to_text
        """

        # no exists path
        mock_os.path.exists.return_value = False
        bad_result_noexistpath = pdf_to_text()
        self.assertEqual(bad_result_noexistpath, [])
        self.assertEqual(mock_open.call_count, 0)

        # pdfminer exception
        mock_os.path.exists.return_value = True
        mock_pdfminer_extract = mock_pdfminer_highlevel.extract_text_to_fp
        mock_pdfminer_extract.side_effect = Exception("Test exception")
        bad_result_exception = pdf_to_text()
        self.assertEqual(bad_result_exception, [])
        self.assertEqual(mock_open.call_count, 1)

        mock_open.reset_mock()
        mock_pdfminer_highlevel.extract_text_to_fp.side_effect = None
        mock_stringio().getvalue.return_value = self.test_rows
        result = pdf_to_text('testfilepath')
        self.assertEqual(result, self.test_rows)
        self.assertEqual(mock_open.call_count, 1)

    def test_pdf_row_limiter(self):
        """
        Test pdf_row_limiter
        """
        no_limiter_results = pdf_row_limiter(self.test_rows)
        self.assertEqual(no_limiter_results, self.test_rows)

        x_limiter_results = pdf_row_limiter(self.test_rows, [2])
        self.assertEqual(x_limiter_results, self.test_rows[2:])

        xy_limiter_results = pdf_row_limiter(self.test_rows, [2, 5])
        self.assertEqual(xy_limiter_results, self.test_rows[2:5])

    def test_pdf_row_format(self):
        """
        Test pdf_row_format
        """
        result = pdf_row_format(self.test_rows[0])
        self.assertEqual(result, self.test_rows[0])

    def test_pdf_row_parser(self):
        """
        Test pdf_row_parser
        """
        result = pdf_row_parser(self.test_rows)
        self.assertEqual(result, self.test_rows)

    def test_pdf_row_cleaner(self):
        """
        Test pdf_row_cleaner
        """
        result = pdf_row_cleaner(['test', '', [], 'test2'])
        self.assertEqual(result, ['test', 'test2'])

    @mock.patch('scrapbag.pdf.pdf_to_text')
    @mock.patch('scrapbag.pdf.pdf_row_format')
    @mock.patch('scrapbag.pdf.pdf_row_limiter')
    @mock.patch('scrapbag.pdf.pdf_row_parser')
    @mock.patch('scrapbag.pdf.pdf_row_cleaner')
    def test_pdf_to_dict(self, mock_pdf_row_cleaner, mock_pdf_row_parser,
                         mock_pdf_row_limiter, mock_pdf_row_format,
                         mock_pdf_to_text):
        """
        Test pdf_to_dict
        """
        pdf_to_dict('testpdffilepath')
        pdf_to_dict('testpdffilepath', rows=self.test_rows)

        # only runs when no rows passed
        self.assertEqual(mock_pdf_to_text.call_count, 1)
        self.assertEqual(mock_pdf_row_format.call_count, 1)

        self.assertEqual(mock_pdf_row_limiter.call_count, 2)
        self.assertEqual(mock_pdf_row_parser.call_count, 2)
        self.assertEqual(mock_pdf_row_cleaner.call_count, 2)



# -*- coding: utf-8 -*-
"""
Mock csv classes module
"""


class MockCell():
    """
    Class to mock excel cells..
    """
    def __init__(self, value):
        self.value = value


class MockSheet():
    """
    Class to mock excel cells..
    """

    def __init__(self, cells, merged_cells=None):

        self.cells = [[MockCell(xcell) for xcell in ycell] for ycell in cells]
        self.ncols = max([len(row) for row in self.cells])
        self.nrows = len(self.cells)
        self.merged_cells = [] if merged_cells is None else merged_cells

    def cell(self, x_cell, y_cell):
        """
        Cell getter
        Returns: value in the list corresponding to the parameters introduced.
        """
        return self.cells[x_cell][y_cell]



# -*- coding: utf-8 -*-
"""
Test Scrapbag csv file
"""
import os
import unittest
import mock


from scrapbag.tests.files.mock_csv import MockSheet, MockCell
from scrapbag.csvs import (
    get_csv_col_headers,
    row_headers_count,
    get_row_headers,
    csv_tolist,
    excel_todictlist,
    search_mergedcell_value,
    row_csv_limiter,
    row_iter_limiter,
    csv_format,
    csv_to_dict,
    excel_to_dict
    )


UTILS_PATH = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))


class UtilsCSVTestCase(unittest.TestCase):
    """
    Scrapbag csv Test Case
    """

    # TODO: PUT THIS IN A TEST AUX FILE..
    test_rows = [
        ['name ', 'Explotaciones. Número', 'Explotaciones. %',
         'Superficie total. Ha.', 'Superficie total.%', 'SAU. Ha.',
         'SAU. %', ''],
        ['Explotaciones con tierras', '51591.0', '100.0', '3044707.0', '100.0',
         '2345696.0', '100.0', ''],
        ['Explotaciones sin SAU', '294.0', '0.57', '2345.0', '0.08', '0.0',
         '0.0', ''],
        ['Explotaciones con SAU', '51297.0', '99.43', '3042361.0', '99.92',
         '2345696.0', '100.0', '']
    ]

    def test_get_csv_col_headers(self):
        """
        Test get_csv_col_headers
        """

        result = get_csv_col_headers(self.test_rows)
        result2 = get_csv_col_headers(self.test_rows, 1)
        result3 = get_csv_col_headers([])

        self.assertEqual(len(result[0]), 8)
        self.assertEqual(len(result2[0]), 7)
        self.assertEqual(result3, [])

    def test_count_row_headers(self):
        """
        Test count_row_headers
        """

        count1 = row_headers_count(self.test_rows)

        self.test_rows[0][0] = ''
        self.test_rows[0][1] = ''

        count2 = row_headers_count(self.test_rows)
        count3 = row_headers_count([])
        count4 = row_headers_count([[]])

        self.assertEqual(count1, 0)
        self.assertEqual(count2, 2)
        self.assertEqual(count3, 0)
        self.assertEqual(count4, 0)

    def test_get_row_headers(self):
        """
        Test get_row_headers
        """

        test_rows = [
            ['', '', 'c1', 'c2', 'c3'],
            ['r1', 'r1a', 'a1', 'a2'],
            ['', 'r1b', 'b1', 'b2'],
            ['r1', 'r2b', 'b3', 'b4'],
            ['', 'r3b', 'b5', 'b6'],
            ['r2', 'r2c', 'c1', 'c2'],
            ['r3', 'r3d', 'd1', 'd2'],
            ['', '', 'd3', 'd4'],
        ]

        expected_rheaders = ['r1 r1a', 'r1 r1b', 'r2 r2c', 'r3 r3d']

        result = get_row_headers(test_rows, 2, 1)
        self.assertEqual(len(result), 7)
        self.assertEqual(
            all([x in result for x in expected_rheaders]), True)

    def test_retrieve_csv_data(self):
        """
        Test retrieve_csv_data
        """
        # TODO: Make test
        pass

    @mock.patch('scrapbag.csvs.open', create=True)
    @mock.patch('scrapbag.files.zipfile.ZipFile')
    @mock.patch('scrapbag.csvs.io.TextIOWrapper')
    @mock.patch('scrapbag.csvs.itertools')
    def test_csv_tolist(self,
                        mock_itertools,
                        mock_io,
                        mock_zipfile,
                        mock_open):

        """
        Test csv_tolist
        """

        mock_open.return_value = True
        mock_itertools.return_value.chain.return_value = [1, 2, 3]
        mock_io.return_value = True
        mock_zipfile.return_value.configure_mock(**{
            'namelist.return_value': ['file1.csv', 'file2.csv'],
            'open.return_value': True,
        })

        result = csv_tolist("scrapman/tests/file.zip")
        self.assertTrue(isinstance(result, list))

        mock_zipfile.return_value.configure_mock(**{
            'namelist.return_value': [],
            'open.return_value': True,
        })

        result2 = csv_tolist("scrapman/tests/file.zip")
        self.assertEqual(result2, [])

    def test_excel_todictlist(self):
        """
        Test excel_todictlist
        """
        path = os.path.join(UTILS_PATH, 'tests/files/xlsx')
        xlsx_testfile = os.path.join(path, 'xlsx_test1.xlsx')
        xlsx_testfile2_with_merges = os.path.join(path, 'xlsx_test2.xlsx')

        result = excel_todictlist(xlsx_testfile)
        result2 = excel_todictlist(xlsx_testfile2_with_merges)

        self.assertTrue(isinstance(result, dict))
        self.assertEqual(len(result.keys()), 2)

        self.assertIn('test_Sheet1', result)
        self.assertTrue(isinstance(result['test_Sheet1'], list))
        self.assertIn('test_Sheet2', result)

        self.assertTrue(isinstance(result2, dict))
        self.assertEqual(len(result.keys()), 2)
        self.assertEqual(len(result['test_Sheet2'][-1]), 4)

        self.assertIn(['', 'x', 'y', 'z', 'z'], result2['test_Sheet2'])
        self.assertEqual(len(result2['test_Sheet2'][-1]), 5)

    def test_search_mergedcell_value(self):
        """
        Test search_mergedcell_value
        """
        test = [[1, 2, 3], [4, "", ""], [7, "", ""]]
        mock_cell = MockSheet(test)

        # find 4
        self.assertEqual(isinstance(search_mergedcell_value(
            mock_cell, (1, 2, 0, 2)), MockCell), True)

        # not found anything
        self.assertEqual(isinstance(search_mergedcell_value(
            mock_cell, (1, 2, 1, 2)), MockCell), False)

        self.assertEqual(search_mergedcell_value(
            mock_cell, (0, 0, 0, 2)), False)
        self.assertEqual(search_mergedcell_value(
            mock_cell, (1, 2, 0, 0)), False)

    def test_populate_headers(self):
        """
        Test populate_headers
        """
        # TODO: Make test
        pass

    def test_row_csv_limiter(self):
        """
        Test row_csv_limiter
        """

        result = row_csv_limiter(self.test_rows)
        result2 = row_csv_limiter(self.test_rows, [])
        result3 = row_csv_limiter(self.test_rows, [2])
        result4 = row_csv_limiter(self.test_rows, [0, 2])

        self.assertEqual(result, self.test_rows)
        self.assertEqual(result2, self.test_rows)
        self.assertEqual(result3, self.test_rows[-2:])
        self.assertEqual(result4, self.test_rows[:2])

    def test_row_iter_limiter(self):
        """
        Test row_iter_limiter
        """
        self.assertEqual(row_iter_limiter([], 0, 1, 0), None)
        self.assertEqual(row_iter_limiter(
            [[1], [1, 3, 4], [4, 5, 6]], 0, 1, 0), 1)

    def test_csv_format(self):
        """
        Test csv_format
        """
        test_data = [[1, 2, 3], [4, 5, 6], [7, 8, 9]]

        result = csv_format(
            test_data, ['c1', 'c2', 'c3'], ['r1', 'r2', 'r3'], result_format=1)
        result_strange = csv_format(test_data, result_format=1)
        result_dict = csv_format(
            test_data, ['c1', 'c2', 'c3'], ['r1', '', 'r3'], result_format=2)
        result_bad_resultformat = csv_format(
            test_data, ['c1', 'c2', 'c3'], ['r1', '', 'r3'], result_format=99)

        self.assertEqual(result_strange, [[[]] + test_data])
        self.assertEqual(
            result, [[['', 'c1', 'c2', 'c3'],
                      ['r1', 1, 2, 3],
                      ['r2', 4, 5, 6],
                      ['r3', 7, 8, 9]]])
        self.assertEqual(isinstance(result_dict, dict), True)
        self.assertEqual(len(result_dict.keys()), 2)
        self.assertEqual(result_bad_resultformat, None)

    @mock.patch('scrapbag.csvs.csv_tolist')
    def test_csv_to_dict(self, mock_csv_tolist):
        """
        Test csv_to_dict
        """
        path = os.path.join(UTILS_PATH, 'tests/files/csv')

        mock_csv_tolist.return_value = []
        with self.assertRaises(ValueError):
            csv_to_dict(os.path.join(path, 'csv_test.csv'), result_format=2)

        mock_csv_tolist.side_effect = csv_tolist
        result = csv_to_dict(os.path.join(
            path, 'csv_test.csv'), result_format=2)[0]
        result_array = csv_to_dict(os.path.join(
            path, 'csv_test.csv'), result_format=0)[0]
        result_array_clean = csv_to_dict(os.path.join(
            path, 'csv_test.csv'), result_format=1)[0]

        # perfect square csv test as dict result format
        self.assertTrue(isinstance(result, list))
        self.assertTrue(isinstance(result[0], dict))
        self.assertEqual(len(result), 12)
        self.assertEqual(len(result[0].keys()), 7)

        # perfect square csv test as array result format
        self.assertTrue(isinstance(result_array, list))
        self.assertTrue(isinstance(result_array[0], list))
        # row header present in first pos of the list
        self.assertEqual(len(result_array), 13)
        self.assertEqual(len(result_array[0]), 7)  # last space not cleaned

        # perfect square csv test as clean array result format
        self.assertTrue(isinstance(result_array_clean, list))
        self.assertTrue(isinstance(result_array_clean[0], list))
        # row header present in first pos of the list
        self.assertEqual(len(result_array_clean), 13)
        # last space not cleaned
        self.assertEqual(len(result_array_clean[0]), 7)

        # perfect square bigger csv test
        result2 = csv_to_dict(os.path.join(
            path, 'csv_test2.csv'), result_format=2)[0]
        self.assertTrue(isinstance(result2, list))
        self.assertTrue(isinstance(result2[0], dict))
        self.assertEqual(len(result2), 396)
        self.assertEqual(len(result[0].keys()), 7)

        # one row header, one column header, changing encoding and dirty rows
        # to limit
        result3 = csv_to_dict(
            os.path.join(path, 'csv_test3.csv'),
            result_format=2,
            encoding='ISO-8859-1',
            limits=[5, -5])

        self.assertTrue(isinstance(result3, dict))
        self.assertTrue(isinstance(result3[list(result3)[0]], dict))
        self.assertEqual(len(result3.keys()), 7)

        # 2 row header, 2 column header
        result4 = csv_to_dict(os.path.join(
            path, 'csv_test4.csv'), result_format=2)
        self.assertTrue(isinstance(result4, dict))
        self.assertTrue(isinstance(result4[list(result4)[0]], dict))
        self.assertEqual(len(result4.keys()), 6)

        # strange row headers
        result6 = csv_to_dict(os.path.join(
            path, 'csv_test5.csv'), result_format=2)
        self.assertTrue(isinstance(result6, dict))
        self.assertTrue(isinstance(result6[list(result6)[0]], dict))
        self.assertEqual(len(result6.keys()), 9)

        # one row only
        result7 = csv_to_dict('', rows=[['a', 'b', 'c']], result_format=2)
        self.assertEqual(result7, [[]])

    @mock.patch('scrapbag.csvs.csv_to_dict')
    @mock.patch('scrapbag.csvs.excel_todictlist')
    def test_excel_to_dict(self, mock_excel_todictlist, mock_csv_to_dict):
        """
        Test excel_to_dict
        """
        path = os.path.join(UTILS_PATH, 'tests/files/xlsx')

        mock_excel_todictlist.side_effect = excel_todictlist
        mock_csv_to_dict.side_effect = csv_to_dict

        # xlsx file
        result5_encapsulate = excel_to_dict(
            os.path.join(path, 'xlsx_test1.xlsx'),
            encapsulate_filepath=True, result_format=2)
        result5 = excel_to_dict(
            os.path.join(path, 'xlsx_test1.xlsx'),
            encapsulate_filepath=False, result_format=2)

        path_encapsulated = os.path.join(path, 'xlsx_test1.xlsx')
        self.assertIn(path_encapsulated, result5_encapsulate)
        self.assertEqual(
            len(result5_encapsulate.get(path_encapsulated, {}).keys()), 2)

        self.assertNotIn(path_encapsulated, result5)
        self.assertEqual(len(result5.keys()), 2)

        # xlsx file merged cells
        result7 = excel_to_dict(
            os.path.join(path, 'xlsx_test2.xlsx'),
            encapsulate_filepath=True, result_format=2)
        path_encapsulated = os.path.join(path, 'xlsx_test2.xlsx')
        self.assertIn(path_encapsulated, result7)

        # xlsx file merged cells
        result8 = excel_to_dict(
            os.path.join(path, 'xlsx_test2b.xls'),
            encapsulate_filepath=True, result_format=2)

        path_encapsulated = os.path.join(path, 'xlsx_test2b.xls')
        self.assertIn(path_encapsulated, result8)

        self.assertEqual(len(result8.get(path_encapsulated, {}).keys()), 3)
        self.assertEqual(result8.get(path_encapsulated, {}).get(
            'test_Sheet3', {}).get('a', {}).get('a-aa', {}), 'vaaa')
        self.assertEqual(result8.get(path_encapsulated, {}).get(
            'test_Sheet3', {}).get('b', {}).get('b-bb', {}), 'vbbb')
        self.assertEqual(result8.get(path_encapsulated, {}).get(
            'test_Sheet3', {}).get('c', {}).get('c-cb', {}), 'vcbc')

        # Excel Exceptions
        mock_csv_to_dict.side_effect = [
            csv_to_dict, Exception('Fail test exception'), csv_to_dict]
        result8_sheet_exception = excel_to_dict(
            os.path.join(path, 'xlsx_test2b.xls'),
            encapsulate_filepath=True, result_format=2)

        self.assertEqual(result8_sheet_exception.get(
            path_encapsulated, {}).get('test_Sheet2', 'not_found_key'), [])

        self.assertEqual(
            result8_sheet_exception.get(
                path_encapsulated, {}).get('test_Sheet1', 'not_found_key'),
            csv_to_dict)

        self.assertEqual(
            result8_sheet_exception.get(
                path_encapsulated, {}).get('test_Sheet3', 'not_found_key'),
            csv_to_dict)

        mock_excel_todictlist.side_effect = Exception('Err')
        result8_exception = excel_to_dict(
            os.path.join(path, 'xlsx_test2b.xls'),
            encapsulate_filepath=True, result_format=2)
        self.assertEqual(result8_exception, {})



# -*- coding: utf-8 -*-
"""
Scrapbag collections file.
"""
import re
import collections
from functools import reduce
from scrapbag.strings import exclude_chars

import structlog

logger = structlog.getLogger(__name__)


def exclude_values(values, args):
    """
    Exclude data with specific value.
    =============   =============   =======================================
    Parameter       Type            Description
    =============   =============   =======================================
    values          list            values where exclude elements
    args            list or dict    elements to exclude
    =============   =============   =======================================
    Returns: vakues without excluded elements
    """

    if isinstance(args, dict):
        return {
            key: value
            for key, value in (
                (k, exclude_values(values, v)) for (k, v) in args.items())
            if value not in values
        }
    elif isinstance(args, list):
        return [
            item
            for item in [exclude_values(values, i) for i in args]
            if item not in values
        ]

    return args


def exclude_empty_values(args):
    """
    Exclude None, empty strings and empty lists using exclude_values.
    =============   =============   =======================================
    Parameter       Type            Description
    =============   =============   =======================================
    args            list or dict    elements to exclude
    =============   =============   =======================================
    Returns: values without excluded values introduces and without defined
    empty values.
    """
    empty_values = ['', 'None', None, [], {}]
    return exclude_values(empty_values, args)


def check_fields(fields, args):
    """Check that every field given in fields is included in args.args.

    - fields (tuple): fieldes to be searched in args
    - args (dict): dictionary whose keys will be checked against fields
    """
    for field in fields:
        if field not in args:
            return False
    return True


def get_element(source, path, separator=r'[/.]'):
    """
    Given a dict and path '/' or '.' separated. Digs into de dict to retrieve
    the specified element.

    Args:
        source (dict): set of nested objects in which the data will be searched
        path (string): '/' or '.' string with attribute names
    """
    return _get_element_by_names(source, re.split(separator, path))


def _get_element_by_names(source, names):
    """
    Given a dict and path '/' or '.' separated. Digs into de dict to retrieve
    the specified element.

    Args:
        source (dict): set of nested objects in which the data will be searched
        path (list): list of attribute names
    """

    if source is None:
        return source

    else:
        if names:
            head, *rest = names
            if isinstance(source, dict) and head in source:
                return _get_element_by_names(source[head], rest)
            elif isinstance(source, list) and head.isdigit():
                return _get_element_by_names(source[int(head)], rest)
            elif not names[0]:
                pass
            else:
                source = None
        return source


def add_element(source, path, value, separator=r'[/.]', **kwargs):
    """
    Add element into a list or dict easily using a path.
    =============   =============   =======================================
    Parameter       Type            Description
    =============   =============   =======================================
    source          list or dict    element where add the value.
    path            string          path to add the value in element.
    value           ¿all?           value to add in source.
    separator       regex string    Regexp to divide the path.
    =============   =============   =======================================
    Returns: source with added value
    """

    return _add_element_by_names(
        source,
        exclude_empty_values(re.split(separator, path)),
        value,
        **kwargs)


def _add_element_by_names(src, names, value, override=False, digit=True):
    """
    Internal method recursive to Add element into a list or dict easily using
    a path.
    =============   =============   =======================================
    Parameter       Type            Description
    =============   =============   =======================================
    src             list or dict    element where add the value.
    names           list            list with names to navigate in src.
    value           ¿all?           value to add in src.
    override        boolean         Override the value in path src.
    =============   =============   =======================================
    Returns: src with added value
    """

    if src is None:
        return False

    else:

        if names and names[0]:
            head, *rest = names

            # list and digit head
            if isinstance(src, list):
                if force_list(digit)[0] and head.isdigit():
                    head = int(head)

                    # if src is a list and lenght <= head
                    if len(src) <= head:
                        src.extend([""] * (head + 1 - len(src)))

            # head not in src :(
            elif isinstance(src, dict):
                if head not in src:
                    src[head] = [""] * (int(rest[0]) + 1) if rest and force_list(digit)[0] and rest[0].isdigit() else {}

            # more heads in rest
            if rest:

                # Head find but isn't a dict or list to navigate for it.
                if not isinstance(src[head], (dict, list)):

                    # creative mode.
                    if override:
                        # only could be str for dict or int for list
                        src[head] = [""] * (int(rest[0]) + 1) if force_list(digit)[0] and rest[0].isdigit() else {}

                        digit = digit if not digit or not isinstance(digit, list) else digit[1:]

                        if not force_list(digit)[0] and rest and str(rest[0]).isdigit() and isinstance(src[head], list) and override:
                            src[head] = {}

                        _add_element_by_names(src[head], rest, value, override=override, digit=digit)

                    else:
                        return False

                else:

                    digit = digit if not digit or not isinstance(digit, list) else digit[1:]

                    if not force_list(digit)[0] and rest and str(rest[0]).isdigit() and isinstance(src[head], list) and override:
                        src[head] = {}

                    _add_element_by_names(src[head], rest, value, override=override, digit=digit)

            # it's final head
            else:

                if not override and isinstance(src[head], list):
                    src[head].append(value)

                elif ((not override and isinstance(src[head], dict)) and
                      (isinstance(value, dict))):
                    src[head].update(value)

                else:
                    src[head] = value

        return src


def format_dict(dic, format_list, separator=',', default_value=str):
    """
    Format dict to string passing a list of keys as order
    Args:
        lista: List with elements to clean duplicates.
    """

    dic = collections.defaultdict(default_value, dic)
    return separator.join([
        "{" + "{}".format(head) + "}" for head in format_list]).format(**dic)


def force_list(element):
    """
    Given an element or a list, concatenates every element and clean it to
    create a full text
    """
    if element is None:
        return []

    if isinstance(element, (collections.Iterator, list)):
        return element

    return [element]


def flatten(data, parent_key='', sep='_'):
    """
    Transform dictionary multilevel values to one level dict, concatenating
    the keys with sep between them.
    """
    items = []

    if isinstance(data, list):
        logger.debug('Flattening list {}'.format(data))
        list_keys = [str(i) for i in range(0, len(data))]
        items.extend(
            flatten(dict(zip(list_keys, data)), parent_key, sep=sep).items())

    elif isinstance(data, dict):
        logger.debug('Flattening dict {}'.format(data))

        for key, value in data.items():
            new_key = parent_key + sep + key if parent_key else key
            if isinstance(value, collections.MutableMapping):
                items.extend(flatten(value, new_key, sep=sep).items())
            else:
                if isinstance(value, list):
                    list_keys = [str(i) for i in range(0, len(value))]
                    items.extend(
                        flatten(
                            dict(zip(list_keys, value)), new_key, sep=sep).items())
                else:
                    items.append((new_key, value))
    else:
        logger.debug('Nothing to flatten with {}'.format(data))
        return data

    return collections.OrderedDict(items)


def nested_dict_to_list(path, dic, exclusion=None):
    """
    Transform nested dict to list
    """
    result = []
    exclusion = ['__self'] if exclusion is None else exclusion

    for key, value in dic.items():

        if not any([exclude in key for exclude in exclusion]):
            if isinstance(value, dict):
                aux = path + key + "/"
                result.extend(nested_dict_to_list(aux, value))
            else:
                if path.endswith("/"):
                    path = path[:-1]

                result.append([path, key, value])

    return result


def find_value_in_object(attr, obj):
    """Return values for any key coincidence with attr in obj or any other
    nested dict.
    """

    # Carry on inspecting inside the list or tuple
    if isinstance(obj, (collections.Iterator, list)):
        for item in obj:
            yield from find_value_in_object(attr, item)

    # Final object (dict or entity) inspect inside
    elif isinstance(obj, collections.Mapping):

        # If result is found, inspect inside and return inner results
        if attr in obj:

            # If it is iterable, just return the inner elements (avoid nested
            # lists)
            if isinstance(obj[attr], (collections.Iterator, list)):
                for item in obj[attr]:
                    yield item

            # If not, return just the objects
            else:
                yield obj[attr]

        # Carry on inspecting inside the object
        for item in obj.values():
            if item:
                yield from find_value_in_object(attr, item)


def remove_list_duplicates(lista, unique=False):
    """
    Remove duplicated elements in a list.
    Args:
        lista: List with elements to clean duplicates.
    """
    result = []
    allready = []

    for elem in lista:
        if elem not in result:
            result.append(elem)
        else:
            allready.append(elem)

    if unique:
        for elem in allready:
            result = list(filter((elem).__ne__, result))

    return result


def dict2orderedlist(dic, order_list, default='', **kwargs):
    """
    Return a list with dict values ordered by a list of key passed in args.
    """
    result = []
    for key_order in order_list:
        value = get_element(dic, key_order, **kwargs)
        result.append(value if value is not None else default)
    return result


def get_dimension(data):
    """
    Get dimension of the data passed by argument independently if it's an
    arrays or dictionaries
    """
    result = [0, 0]

    if isinstance(data, list):
        result = get_dimension_array(data)

    elif isinstance(data, dict):
        result = get_dimension_dict(data)

    return result


def get_dimension_dict(dictionary):
    """
    Get dimension of an dict passed by arg

    Arguments:
        dictionary {[dict]} -- [dict passed to take dimension]

    Returns:
        [list] -- [dictionary dimension size]
    """
    result = [1, len(dictionary)]
    return result


def get_dimension_array(array):
    """
    Get dimension of an array getting the number of rows and the max num of
    columns.
    """
    if all(isinstance(el, list) for el in array):
        result = [len(array), len(max([x for x in array], key=len,))]

    # elif array and isinstance(array, list):
    else:
        result = [len(array), 1]

    return result


def get_ldict_keys(ldict, flatten_keys=False, **kwargs):
    """
    Get first level keys from a list of dicts
    """
    result = []
    for ddict in ldict:
        if isinstance(ddict, dict):

            if flatten_keys:
                ddict = flatten(ddict, **kwargs)

            result.extend(ddict.keys())
    return list(set(result))


def get_alldictkeys(ddict, parent=None):
    """
    Get all keys in a dict
    """
    parent = [] if parent is None else parent

    if not isinstance(ddict, dict):
        return [tuple(parent)]
    return reduce(
        list.__add__,
        [get_alldictkeys(v, parent + [k]) for k, v in ddict.items()],
        [])


def clean_dictkeys(ddict, exclusions=None):
    """
    Exclude chars in dict keys and return a clean dictionary.
    """
    exclusions = [] if exclusions is None else exclusions

    if not isinstance(ddict, dict):
        return {}

    for key in list(ddict.keys()):
        if [incl for incl in exclusions if incl in key]:
            data = ddict.pop(key)
            clean_key = exclude_chars(key, exclusions)

            if clean_key:
                if clean_key in ddict:
                    ddict[clean_key] = force_list(ddict[clean_key])
                    add_element(ddict, clean_key, data)
                else:
                    ddict[clean_key] = data

        # dict case
        if isinstance(ddict.get(key), dict):
            ddict[key] = clean_dictkeys(ddict[key], exclusions)

        # list case
        elif isinstance(ddict.get(key), list):
            for row in ddict[key]:
                if isinstance(row, dict):
                    row = clean_dictkeys(row, exclusions)

    return ddict



# -*- coding: utf-8 -*-
"""
Scrapbag spss file.
"""
import re
import codecs
import structlog

from .collections import exclude_empty_values, add_element, force_list
from .strings import get_only_words

logger = structlog.getLogger(__name__)


def parse_spss_header_leyend(raw_leyend, leyend={}, **kwargs):

    try:
        for raw_leyend_line in raw_leyend.split('/')[1].split('\r\n'):
            head = []

            for raw_value in raw_leyend_line.strip().split():

                clean_value = get_only_words(re.sub(r'\(.*\)', '', raw_value))

                if all([x.isdigit() for x in clean_value]):
                    str_head = " ".join(head)

                    # supossed that only questions uses TO
                    if ' TO ' in str_head:
                        id_range = list(map(int, re.findall(r'(\d+)', str_head)))
                        id_range[1] += 1

                        column_range = list(map(int, clean_value))
                        column_range[1] += 1

                        for i, column in zip(range(*id_range, 1), range(*column_range, 1)):
                            add_element(
                                leyend,
                                '{}/columns'.format("P{}".format(i)),
                                [column - 1, int(column)])

                    # More than one head waiting to assign leyend columns
                    elif len(head) > 1:

                        num_heads = len(head)

                        column_range = list(map(int, clean_value))
                        column_range[1] += 1
                        non_iter_column_range = [x for x in range(*column_range)]

                        offset = int(len(non_iter_column_range) / num_heads)

                        for head_i, column in zip(iter(head), range(*column_range, offset)):

                            add_element(
                                leyend,
                                '{}/columns'.format(head_i),
                                [column - 1, column + offset - 1] if offset != 1 else [column - 1, column])

                    # only one head
                    else:

                        list_clean_value = list(map(int, clean_value))

                        if len(list_clean_value) == 1:
                            list_clean_value.append(list_clean_value[0])

                        list_clean_value[0] -= 1

                        add_element(
                            leyend,
                            '{}/columns'.format(" ".join(head)),
                            list_clean_value
                            )

                    head = []

                else:
                    head.append(raw_value)

    except Exception as ex:
        logger.error('Fail to parse leyend - {}'.format(ex))

    return leyend


def parse_variable_label(raw_labels, headers):

    for label in [exclude_empty_values(x.strip().split("'")) for x in raw_labels]:
        try:
            add_element(
                headers,
                '{}/variable_label'.format(label[0].replace('/', '').strip()),
                label[1].strip())

        except Exception as ex:
            logger.error('Fail to parse variable label', label=label)


def parse_value_label(raw_labels, headers):

    heads = []
    raw_labels[0] = '/' + raw_labels[0].strip()

    for label in raw_labels:
        try:

            if '/' in label.strip()[0]:
                heads = label.replace('/', '').strip().split()

            else:
                for label_value in exclude_empty_values(label.split("' ")):
                    try:

                        v_name, *v_value = label_value.split("'")

                        for head in force_list(heads):

                            add_element(
                                headers,
                                '{}/value_label'.format(head.strip()),
                                {v_name.strip(): " ".join(force_list(v_value)).strip()},
                                digit=False)

                    except Exception as ex:
                        logger.error('Fail to parse label value {}'.format(ex), label_value=label_value)
                        continue

        except Exception as ex:
            logger.error('Fail to parse variable label - {}'.format(ex), label=label)


def parse_spss_header_labels(raw_labels, headers={}, **kwargs):
    try:
        raw_labels_splited = exclude_empty_values(raw_labels.split('\r\n'))

        if 'value' in raw_labels_splited[0].lower():
            parse_value_label(raw_labels_splited[1:], headers)

        else:
            parse_variable_label(raw_labels_splited[1:], headers)

    except Exception as ex:
        logger.error('Fail to parse leyend - {}'.format(ex))

    return headers


def parse_spss_headerfile(path, **kwargs):
    """
    Parse spss header file

    Arguments:
        path {str} -- path al fichero de cabecera.
        leyend_position -- posicion del la leyenda en el header.
    """
    headers_clean = {}
    try:
        with codecs.open(path, 'r', kwargs.get('encoding', 'latin-1')) as file_:
            raw_file = file_.read()
            raw_splited = exclude_empty_values(raw_file.split('.\r\n'))

            # suposse that by default spss leyend is in position 0.
            leyend = parse_spss_header_leyend(
                raw_leyend=raw_splited.pop(kwargs.get('leyend_position', 0)),
                leyend=headers_clean)

            if not leyend:
                raise Exception('Empty leyend')

            # only want VARIABLE(S) LABEL(S) & VALUE(S) LABEL(S)
            for label in [x for x in raw_splited if 'label' in x.lower()]:
                values = parse_spss_header_labels(
                    raw_labels=label,
                    headers=leyend)

    except Exception as ex:
        logger.error('Fail to parse spss headerfile - {}'.format(ex), header_file=path)
        headers_clean = {}

    return headers_clean


def parse_spss_datafile(path, **kwargs):
    """
    Parse spss data file

    Arguments:
        path {str} -- path al fichero de cabecera.
        **kwargs {[dict]} -- otros argumentos que puedan llegar
    """
    data_clean = []
    with codecs.open(path, 'r', kwargs.get('encoding', 'latin-1')) as file_:
        raw_file = file_.read()
        data_clean = raw_file.split('\r\n')
    return exclude_empty_values(data_clean)


def join_spss_header_data(header, data, question_dict={}, **kwargs):
    joined_results = []

    # lenght of data
    num_data = len(data)

    for i, data_raw in enumerate(data):
        try:
            if i % 50 == 0:
                logger.debug('Joining spss header with data {} / {}'.format(i, num_data))

            partial_results = {}
            for key, header_data in header.items():
                try:
                    columns_raw = header_data.get('columns')

                    if not columns_raw:
                        continue

                    columns = slice(*columns_raw)

                    response_raw = data_raw[columns].strip()
                    response = header_data.get('value_label', {}).get(response_raw) or header_data.get('value_label', {}).get('0' + response_raw) or response_raw

                    partial_results[key] = {
                        'question': question_dict.get(key),
                        'variable_label': header_data.get('variable_label'),
                        'response': response.strip()
                    }
                except Exception as ex:
                    logger.error("Fail to parse header key".format(ex), key=key)
                    continue

        except Exception as ex:
            logger.error("".format('Fail to parse data_raw - {}'.format(ex)))
            continue

        joined_results.append(exclude_empty_values(partial_results))
    return joined_results


def parse_spss(header_filepath, data_filepath, **kwargs):

    result = {}

    try:
        # parse header
        header = parse_spss_headerfile(header_filepath, **kwargs)

        # parse data
        data = parse_spss_datafile(data_filepath, **kwargs)

        # join header & data
        result = join_spss_header_data(header, data, **kwargs)

    except Exception as ex:
        logger.error('Fail to parse spss - {}'.format(ex), header=header_filepath, data=data_filepath)

    return result



import os
import sys
from setuptools import setup, find_packages
from pip.req import parse_requirements

install_reqs = parse_requirements('requirements.txt', session=False)
required = [str(ir.req) for ir in install_reqs]

_here = os.path.abspath(os.path.dirname(__file__))
sys.path.insert(0, _here)


def version():
    import imp
    path = os.path.join(_here, 'scrapbag', 'version.py')
    mod = imp.load_source('version', path)
    return mod.__version__

setup(
    name='scrapbag',
    packages=find_packages(),
    version=version(),
    description='Scrapbag Python lib',
    author='Datary developers team',
    author_email='support@datary.io',
    url='https://github.com/Datary/scrapbag',
    download_url='https://github.com/Datary/scrapbag',
    keywords=[
        'scrapman',
        'scrapbag',
        'utils',
        'datary',
        'scrapping'
    ],
    classifiers=[
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5'
    ],
    install_requires=required,
)



# -*- coding: utf-8 -*-
"""
Datary Mock_Csv Test
"""
import unittest
from scrapbag.tests.files.mock_csv import MockSheet, MockCell


class MockSheetTestCase(unittest.TestCase):
    """
    MockSheet Test Case
    """

    def test(self):
        """
        Test MockSheet        """
        test_list = [[1, 2, 3], [4, 5, 6]]
        test_merged_cells = [(0, 1, 1, 2), (3, 4, 1, 2)]
        test = MockSheet(test_list, merged_cells=test_merged_cells)
        self.assertEqual(isinstance(test.cells, list), True)
        self.assertEqual(len(test.cells), len(test_list))
        self.assertEqual(isinstance(test.cell(0, 2), MockCell), True)
        self.assertEqual(test.cell(0, 2).value, 3)
        self.assertEqual(test.ncols, 3)
        self.assertEqual(test.nrows, 2)
        self.assertEqual(test.merged_cells, test_merged_cells)



# -*- coding: utf-8 -*-
"""
Test Scrapbag dates file
"""
import unittest
from datetime import datetime, timedelta
from scrapbag.dates import (
    get_dates_in_period,
    ordinal_suffix,
    localize_date,
    get_year_from_date_str,
    get_month_from_date_str,
    replace_month_abbr_with_num,
    translate_month_abbr,
    merge_datetime
    )


class UtilsDatesTestCase(unittest.TestCase):
    """
    Scrapbag dates Test Case
    """

    def test_get_dates_in_period(self):
        """
        Test get_dates_in_period
        """

        test = get_dates_in_period()

        # default return a list of next 30 days from today
        self.assertEqual(len(test), 30)
        self.assertEqual(test[0].date(), datetime.today().date())
        self.assertEqual(
            test[-1].date(), (datetime.today() + timedelta(days=30-1)).date())

        test2 = get_dates_in_period(
            top=datetime.today() + timedelta(days=6), step=2)

        self.assertEqual(len(test2), 3)

        for index in range(0, 3):
            self.assertEqual(
                test2[index].date(),
                (datetime.today() + timedelta(days=2*index)).date())

    def test_ordinal_suffix(self):
        """
        Test ordinal_suffix
        """

        self.assertEqual(ordinal_suffix(1), 'st')
        self.assertEqual(ordinal_suffix(2), 'nd')
        self.assertEqual(ordinal_suffix(3), 'rd')
        self.assertEqual(ordinal_suffix(4), 'th')
        self.assertEqual(ordinal_suffix(24), 'th')

    def test_localize_date(self):
        """
        Test localize_date
        """

        test_date = datetime.now()

        canarias = localize_date(test_date, 'Atlantic/Canary')
        madrid = localize_date(test_date, 'Europe/Madrid')
        berlin = localize_date(test_date, 'Europe/Berlin')

        self.assertEqual(canarias.tzname() != madrid.tzname(), True)
        self.assertEqual(berlin.tzname(), madrid.tzname())

    def test_get_year_from_date_str(self):
        """
        Test get_year_from_date_str
        """

        result = get_year_from_date_str('marzo 1991 lalal')
        result2 = get_year_from_date_str('marzo1991lalal')
        result_bad = get_year_from_date_str('marzo199lalal')
        result_bad2 = get_year_from_date_str('marzolalal')

        self.assertEqual(result, result2)
        self.assertEqual(result, ['1991'])
        self.assertEqual(result_bad, [])
        self.assertEqual(result_bad, result_bad2)

    def test_get_month_from_date_str(self):
        """
        Test get_month_from_date_str
        """

        self.assertEqual(get_month_from_date_str(
            '5 de marzo de 2017', 'es'), (3, 'mar'))
        self.assertEqual(get_month_from_date_str(
            '5 de march de 2017', 'en'), (3, 'Mar'))
        self.assertEqual(
            get_month_from_date_str('5 de ma123o de 2017', 'en'), ())

    def test_replace_month_abbr_num(self):
        """
        Test replace_month_abbr_num
        """

        self.assertEqual(replace_month_abbr_with_num(
            '10-ene-2013', lang="es"), '10-1-2013')
        self.assertEqual(replace_month_abbr_with_num(
            '11-feb-2012', lang="es"), '11-2-2012')
        self.assertEqual(replace_month_abbr_with_num(
            '12-mar-2011', lang="es"), '12-3-2011')
        self.assertEqual(replace_month_abbr_with_num(
            '13-abr-2010', lang="es"), '13-4-2010')
        self.assertEqual(replace_month_abbr_with_num(
            '14-may-2009', lang="es"), '14-5-2009')
        self.assertEqual(replace_month_abbr_with_num(
            '15-jun-2008', lang="es"), '15-6-2008')
        self.assertEqual(replace_month_abbr_with_num(
            '16-jul-2007', lang="es"), '16-7-2007')
        self.assertEqual(replace_month_abbr_with_num(
            '17-ago-2006', lang="es"), '17-8-2006')
        self.assertEqual(replace_month_abbr_with_num(
            '18-sep-2005', lang="es"), '18-9-2005')
        self.assertEqual(replace_month_abbr_with_num(
            '19-oct-2004', lang="es"), '19-10-2004')
        self.assertEqual(replace_month_abbr_with_num(
            '20-nov-2003', lang="es"), '20-11-2003')
        self.assertEqual(replace_month_abbr_with_num(
            '21-dic-2002', lang="es"), '21-12-2002')
        self.assertEqual(replace_month_abbr_with_num(
            'Apr 23, 2002', lang="en"), '4 23, 2002')

    def test_translate_month_abbr(self):
        """
        Test translate_month_abbr
        """
        test_date_str = '12-ene-2013'

        with self.assertRaises(Exception):
            translate_month_abbr(test_date_str, 'test', 'en')

        with self.assertRaises(Exception):
            translate_month_abbr(test_date_str, 'es', 'test')

        result = translate_month_abbr(test_date_str, 'es', 'en')
        self.assertEqual(result, '12-Jan-2013')
        self.assertEqual(translate_month_abbr(
            result, 'en', 'es'), test_date_str)

    def test_merge_datetime(self):
        """
        Test merge_datetime
        """
        test_datetime = "15/02/2017"
        test_datetime2 = "15-02-2017"
        test_time = "12:56"
        test_time2 = "22:56"

        result = merge_datetime(test_datetime, test_time)
        result2 = merge_datetime(
            test_datetime2, test_time2, date_format='%d-%m-%Y')
        result3 = merge_datetime(test_datetime)

        self.assertEqual(result, datetime(2017, 2, 15, 12, 56))
        self.assertEqual(result2, datetime(2017, 2, 15, 22, 56))
        self.assertEqual(result3, datetime(2017, 2, 15))
